#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "geosim.h"
#include "gspriv.h"

/* Defined in intrface.c */
extern GSWindow *Win;
extern Iconfig InfState;

/*------------------------------------------------------------------------------
  Functions to support help screens, alert boxes and help messages.
  To use these functions, the module programmer must adopt the following
  naming conventions in the interface file

  The help window is named "help".
    It contains buttons named "helpnext" and "helpprev".

  The message window is named "message".
    Three strings exist named "hmx1" "hmx2" and "hmx3" where x is
    the number of the help message.

  The alert window is named "alert".

  The root window is named "root".
    There is a field named "version".
------------------------------------------------------------------------------*/

/* Globals for GSdisplayhelpmessage */
static int Currmsg = -1;
static int Lastmsg = -1;
boolean AlertUp = FALSE;

/*------------------------------------------------------------------------------
  Write a message in the help box at the bottom of the screen.
  GLOBALS: Currmsg, Lastmsg.
------------------------------------------------------------------------------*/
void GSdisplayhelpmessage (
  int whichMessage
)
{
  int      savewin;
  GSString m1;
  GSString m2;
  GSString m3;
  FontSizeType savefont;
  int      savecolor;
  GSString tstr;

  /*--- Save state and restore it at end ---*/
  savewin = InfState.cw;
  savefont = GSPgetfont();
  savecolor = GSgetcolor();

  if (whichMessage == Currmsg)
  { /*--- whichMessage is already displayed, so return ---*/
    return;
  }
  switch (whichMessage)
  {
    case CURR_MSG: /*--- redisplay current message; useful in redraw_func ---*/
      whichMessage = Currmsg;
      break;
    case PREV_MSG: /*--- display the previous message ---*/
      whichMessage = Lastmsg;
      Currmsg = Lastmsg;
      break;
    default: /*--- whichMessage is the number of a message ---*/
      Lastmsg = Currmsg;
      Currmsg = whichMessage;
      break;
  }

  if (whichMessage > 0)     /*--- make sure whichMessage is positive ---*/
  {
    GSredrawnamedwindow("message");
    strcpy(m1, "hm");
    strcpy(m2, "hm");
    strcpy(m3, "hm");
    sprintf(tstr, "%d", whichMessage*10);
    strcat(m1, tstr);
    sprintf(tstr, "%d", whichMessage*10+1);
    strcat(m2, tstr);
    sprintf(tstr, "%d", whichMessage*10+2);
    strcat(m3, tstr);
    GSsetfont(LARGE);
    GSsetcolor(Win[InfState.cw].fgcolor);
    GSwwritenamedmsg(25, 12, m1);
    GSwwritenamedmsg(25, 22, m2);
    GSwwritenamedmsg(25, 32, m3);
  }

  /*--- Restore state saved on function entry ---*/
  GSPsetcurrentwindow(savewin);
  GSsetfont(savefont);
  GSsetcolor(savecolor);
}


/*------------------------------------------------------------------------------
  Bring up a popup window containing an alert message.
  GLOBALS: None.
------------------------------------------------------------------------------*/
void GSalert (
  int   helpmsg,
  int   bstore,
  char *str,
  ...
)
{
  va_list args;
  FontSizeType     savefont;
  int     savecolor;
  int     y = 15;

  savefont = GSPgetfont();
  savecolor = GSgetcolor();

  GSdisplayhelpmessage(helpmsg);
  if (AlertUp)
    GSredrawnamedwindow("alert");
  else
  {
    GSbell();
    GSdrawnamedwindow("alert", bstore, POPUP);
    AlertUp = TRUE;
  }
  GSsetfont(LARGE);
  va_start(args, str);
  GSsetcolor(Win[InfState.cw].fgcolor);
  while (strcmp(str, "") != 0)
  {
    GSwwritemsg(6, y, GSgetmsg (str));
    y += 15;
    str = va_arg (args, char *);
  }
  GSsetfont(savefont);
  GSsetcolor(savecolor);
}


/*------------------------------------------------------------------------------
  Display a tutorial or help file.
  GLOBALS: None.
------------------------------------------------------------------------------*/
boolean GSdohelpscreens(
  BigString      filename,    /* help file name */
  HelpScreenPage pagecommand, /* page to display */
  int            bstore,      /* use bstore or not */
  int            helpmsg      /* help message num */
)
{
  FILE      *fp;              /* file pointer variable */
  BigString  text_line;       /* one line of text */
  int        i;
  int        j;
  char       ch;
  static int pagenum;
  int        pagesize = 25;
  char      *temp_ptr;
  FontSizeType        savefont;
  int        savecolor;


  savefont = GSPgetfont();
  savecolor = GSgetcolor();

  GSdisplayhelpmessage(helpmsg);
  switch (pagecommand)
  {
    case FirstPage: /*--- haven't started looking at a file yet ---*/
      GSdrawnamedwindow("help", bstore, POPUP);
      pagenum = 0;
      break;
    case PrevPage: /*--- going back to previous screen ---*/
      pagenum -= 2;
      if (pagenum < 0)
      {
        pagenum = 0;
      }
      GSredrawnamedwindow("help");
      break;
    case NextPage:
      GSredrawnamedwindow("help");
      break;
    case CurrPage:
      pagenum--;
      if (pagenum < 0)
      {
        pagenum = 0;
      }
      GSredrawnamedwindow("help");
      break;
  }

  if ((fp = GSopen(filename,"r")) == NULL)
  {
    GSdisablebutton("help","helpnext", DRAW);
    GSdisablebutton("help", "helpprev", DRAW);
    GSsetfont(SMALL);
    GSsetcolor(Win[InfState.cw].fgcolor);
    GSwwritemsg(15, 13, "Unable to open this help file");
    GSsetfont(savefont);
    GSsetcolor(savecolor);
    return(FALSE);
  }

  /*--- read in lines previously viewed  ---*/
  for (j = 0; j < pagenum; j++)
  {
    i = 0;
    while (i < pagesize)
    {
      if (fgets(text_line, BIGSTRINGLEN+1, fp) != NULL)
      {
        temp_ptr = strchr(text_line, '\f');
        if (temp_ptr)
        {
          i = pagesize;
        }
      }
      else
      {
        break;
      }
      i++;
    }
  }

  GSsetfont(SMALL);
  GSsetcolor(Win[InfState.cw].fgcolor);
  i = 0;
  while (i < pagesize)
  {
    if (fgets (text_line, BIGSTRINGLEN+1, fp) != NULL)
    {
      GSwwritemsg (15, 13*(i+1), text_line);
      /*--- If there is a formfeed in the line, stop writing this screen ---*/
      temp_ptr = strchr(text_line, '\f');
      if (temp_ptr)
      {
        i = pagesize;
      }
    }
    else
    {
      break;
    }
    i++;
  }

  if (i == pagesize)
  {
    ch = fgetc (fp);
  }
  if (feof(fp) && (pagenum == 0) )
  { /*--- if help file is only one page long, don't draw buttons ---*/
    GSinvisiblebutton("help", "helpnext", DRAW);
    GSinvisiblebutton("help", "helpprev", DRAW);
  }
  else
  {
    if (feof(fp))
    { /*--- if we are at end of help file, disable the Next button ---*/
      GSdisablebutton("help","helpnext", DRAW);
      GSenablebutton("help", "helpprev", DRAW);
    }
    else
    { /*--- otherwise, enable the Next button ---*/
      ungetc (ch, fp);
      GSenablebutton("help", "helpnext", DRAW);
      if (pagenum != 0)
      {
        GSenablebutton("help", "helpprev", DRAW);
      }
      else
      {
        GSdisablebutton("help", "helpprev", DRAW);
      }
    }
  }
  pagenum ++;
  fclose(fp);
  GSsetfont(savefont);
  GSsetcolor(savecolor);
  return(TRUE);
}


/*------------------------------------------------------------------------------
  Display the version number
------------------------------------------------------------------------------*/
void GSdisplayversion (
  BigString versionname
)
{
  static BigString line = { "" };
  FILE            *fptr = 0;
  int              x, y, w, h;
  GSColor          color;
  FontSizeType              savefont;
  int              savecolor;
  int              savewin;

  /*--- read the version number in from a file the first time ---*/
  if (strcmp (line, "") == 0)
  {
    if ((fptr = GSopen (versionname, "r")) == NULL)
    {
      GSelog(0, "Can't open version file |%s|\n", versionname);
    }
    else
    {
      while ((fgets (line, 80, fptr)) != NULL)
      {
        if (strncmp ("Version", line, 7) == 0)
        {
          fclose (fptr);
          break;
        }
      }
    }
  }
  if (strcmp (line, "") != 0)
  {
    /*--- Save state and restore it at end ---*/
    savewin = InfState.cw;
    savefont = GSPgetfont();
    savecolor = GSgetcolor();

    GSgetfieldrect("VERSION", &x, &y, &w, &h, &color);
    GSsetcurrentnamedwindow ("root");
    GSsetfont(SMALL);
    GSsetcolor(color);
    /*GSwwritemsg (x, y+SMALL, line);*/
    GSwwritemsg (x, y+GStextheight(line), line);

    /*--- Restore state saved on function entry ---*/
    GSPsetcurrentwindow(savewin);
    GSsetfont(savefont);
    GSsetcolor(savecolor);
  }
}

/*------------------------------------------------------------------------------
  Returns the Current message number
------------------------------------------------------------------------------*/
int GSgetcurrmsg (
  void
)
{
  return(Currmsg);
}
