/***********************************************************************
GeoSim GENeric database READER functions.

Originally written by Sheryl Kriss.
Rewritten by Cliff Shaffer, 5/1992.
***********************************************************************/
#include <stdio.h>
#include <string.h>

#include "geosim.h"

static FILE *GSDbfp = NULL; /* Pointer to database file.  This will not be
                        accessable to module-specific reader (i.e., all
                        access to the database file will be through generic
                        reader functions). */

/*------------------------------------------------------------------------------
This function is the driver for reading all GeoSim ASCII tagged
databases.  It is given a file pointer to the database, a list of
keywords, a module-specific function (called the "local reader" to
perform action on those keywords, and a pointer to data space used by
the local reader.  GenReader then scans through the database looking
for a match to any of the supplied keywords.  When one is found, the
local reader is called to process the record.  The local reader uses the
library of generic reader I/O routines appearing below to access data
values from the database.  GenReader returns TRUE if successful, FALSE
if there is an error.

(STATIC) GLOBALS ACCESSED: GSDbfp
------------------------------------------------------------------------------*/
boolean GSgenreader (
  FILE    *fp,      /*--- file pointer to database file ---*/
  boolean (*reader_func)(BigString, ReaderStatusType, void *),
                    /*--- local reader function ---*/
  char    *dbident, /*--- identifier tag for expected database ---*/
  void    *data)    /*--- pointer to block of data used to communicate
                          information from the local reader to its calling
                          routine.--- */
{
  int  rc;                      /* return code */
  BigString buffer;             /* holds scanned pieces of data */
  int ch;                       /* used to read and discard comment text */
  boolean ever_called = FALSE;  /* Did local reader ever get called?
                                   If not, data file is garbage. */
  boolean checkDB = FALSE;      /* Has database ident tag been verified? */

  /* In order to support nested calls to GSgenreader, the current global
     state must be saved.  GSDbfp is the sole contents of that state.
     "save_fp" is used to save the previous value. */
  FILE *save_fp = GSDbfp;

  GSDbfp = fp; /* make available to other GS reader functions */
  /* Give the local reader a chance to initialize */
  if (!(*reader_func)(NULL, READ_INIT, data))
  {
    GSDbfp = save_fp;
    return(FALSE);
  }
  if (dbident == NULL)
    checkDB = TRUE;  /* caller doesn't want tag verified */
  do
  {
    /* read a piece of data */
    rc = GSrdword(buffer, BIGSTRINGLEN);
    if(strlen(buffer) > 0)
      if (buffer[0] == COMMENT_SYM)
        /* comment text -- skip rest of this line */
        for(ch = '\0'; (ch != '\n') && (ch != EOF); ch = getc(GSDbfp));
      else if (!checkDB)
      { /* this MUST be the keyword "database=" and the next word
           MUST be dbident. */
        if (GSstricmp(buffer, "database=") != 0)
        {
          GSelog(0, "Not a GeoSim database\n");
          GSDbfp = save_fp;
          return(FALSE);
        }
        if (!GSrdword(buffer, BIGSTRINGLEN))
        { GSDbfp = save_fp; return(FALSE); }
        if (GSstricmp(buffer, dbident) != 0)
        {
          GSelog(0, "Database identification tag was not |%s|\n", dbident);
          GSDbfp = save_fp;
          return(FALSE);
        }
        checkDB = TRUE;
      }
      else if(buffer[strlen(buffer)-1] == '=')
      { /* determine if the scanned data is a keyword - must end in '=' */
        /* call application specific reader function,
           sending it the keyword scanned */
        ever_called = TRUE;
        if(!(*reader_func)(buffer, READ_PROCESS, data))
        { GSDbfp = save_fp; return(FALSE); }
      }
  } while (rc);
  /* Give the local reader a chance to do post processing */
  if(!(*reader_func)(NULL, READ_FINAL, data))
  { GSDbfp = save_fp; return(FALSE); }
  GSDbfp = save_fp;
  return(ever_called);
}


/***********************************************************************
Skip over whitespace and comments.
GLOBALS: GSDbfp
***********************************************************************/
static int skipwaste(void)
{
  int ch; /* declare "int" for getc */

  /* scan through spaces preceeding the string */
  do
    ch = getc(GSDbfp);
  while ((ch == ' ') || (ch == '\n') || (ch == '\t'));

  while (ch == COMMENT_SYM)
  { /* Remove comments */
    do       /*  read in comment to end of line  */
      ch = getc(GSDbfp);
    while (ch != '\n' && ch != '\r' && ch != EOF);
    if (ch == EOF)
      return(ch);
    /* scan through spaces preceeding the string */
    do
      ch = getc(GSDbfp);
    while ((ch == ' ') || (ch == '\n') || (ch == '\t'));
  }  /*  end while loop for removing comments  */
  return(ch);
}


/**********************************************************************
Interprets the next string in the file pointed to by GBDbfp as a
(long) integer, and returns this value.

(STATIC) GLOBALS ACCESSED: GSDbfp
**********************************************************************/
long GSrdinteger(void)
{
  long return_val;

  /* There had better not be a comment preceding the number! */
  if(fscanf(GSDbfp,"%ld", &return_val) == EOF)
    GSelog(0, "Unexpected error in GSrdinteger\n");
  return((long)return_val);
}


/**********************************************************************
Interprets the next string in the file pointed to by GSDbfp as a
float, and returns this value.

(STATIC) GLOBALS ACCESSED: GSDbfp
**********************************************************************/
double GSrdreal(void)
{
  double return_real;

  /* There had better not be a comment preceding the number! */
  if(fscanf(GSDbfp,"%lf",&return_real) == EOF)
    GSelog(0, "Unexpected EOF in GSrdreal\n");
  return(return_real);
}


/* Flags used by rd_stringhelp to deal with spaces,
   flags passed by GSrdword, GSrdphrase and GSrdstringeol. */
#define NOSPACES 1
#define ONESPACE 0
#define STRIPPING -1
/**********************************************************************
Working funtion for GSrdword, GSrdphrase and GSrdstringeol.
Reads and returns a string terminated by '\n', '\t', EOF, COMMENT_SYM.
If "sawspace" is ONESPACE, then strings may contain (single) spaces.

(STATIC) GLOBALS ACCESSED: GSDbfp
**********************************************************************/
static boolean rd_stringhelp (
  char    *string,
  int     size,
  boolean sawspace,
  int     *end_ch_p)
{
  int  pos = 0;   /* position in string */
  int ch; /* current character in file (int for "getchar") */

  ch = skipwaste();
  if (ch == EOF)
  {
    string[0] = '\0';
    return(FALSE);
  }

  string[pos++] = (char)ch;

  for (;;)
  { /* process string */
    ch = getc(GSDbfp);
    if ((ch == ' ') && (sawspace == ONESPACE))
    { /* Use first space in GSrdphrase */
      string[pos++] = (char)ch;
      ch = getc(GSDbfp);
    }
    if (((ch == ' ') || (ch == '\t')) && (sawspace == STRIPPING))
    {
      string[pos++] = ' ';
      while ((ch == ' ') || (ch == '\t'))
        ch = getc(GSDbfp);
    }
    if ((ch == ' ') ||(ch == '\n') || (ch == '\r') || (ch == COMMENT_SYM) ||
	(ch == '\t') || (ch == EOF))
    { /* reached end of the string */
      if(string[pos-1] == ' ') pos--; /* remove last space if any */
      if (end_ch_p)
        *end_ch_p = ch;
      string[pos] = '\0';
      return(TRUE);
    }
    else if (pos >= size)
    {
      string[size] = '\0';
      GSelog(0, "String exceded maximum size in rd_stringhelp\n");
      return(FALSE);
    }
    else
      string[pos++] = (char)ch;
  } /* for (;;) */
}


/**********************************************************************
Reads and returns the rest of the current line (ignoring comments),
compressing multiple spaces to one space.

GLOBALS: none
**********************************************************************/
boolean GSrdstringeol(
  char *string,    /* string returned to the calling procedure */
  int size)        /* maximum number of characters allowed */
{
  return(rd_stringhelp(string, size, STRIPPING, NULL));
}


/**********************************************************************
Reads and returns in "string" a string delimited by
' ' '\n', '\t', EOF, or COMMENT_SYM, with maximum length "size".
Equivalent to "fscanf(GSDbfp, "%s", string) except that this function
can't blow it bounds due to "size" limit.

GLOBALS: none.
**********************************************************************/
boolean GSrdword(
  char *string,  /* string returned to the calling procedure */
  int size)      /* Maximum length for string */
{
  return(rd_stringhelp(string, size, NOSPACES, NULL));
}


/**********************************************************************
Reads and returns a string terminated by '\n', '\t', EOF, COMMENT_SYM,
or multiple spaces.  Maximum string length is "size".

GLOBALS ACCESSED: none.
**********************************************************************/
boolean GSrdphrase(
  char *string,  /* string returned to the calling procedure */
  int size)      /* Maximum length for string */
{
  return(rd_stringhelp(string, size, ONESPACE, NULL));
}


/**********************************************************************
Return a boolean value -- TRUE iff the word read is TRUE (ignoring case).
GLOBALS ACCESSED: none.
**********************************************************************/
boolean GSrdboolean(void)
{
  BigString string;

  if (!GSrdword(string, BIGSTRINGLEN))
    return(FALSE);
  if (GSstricmp(string, "true") == 0)
    return(TRUE);
  else
    return(FALSE);
}


/**********************************************************************
Read the keywords following the "format" keyword and return these in an
array.  Keywords are scanned until the "tamrof" keyword is encountered.
Returns FALSE if there is a premature EOF.

GLOBALS ACCESSED: NONE
**********************************************************************/
boolean GSrdformat(
  int  *numwords,       /* number of format keywords read */
  FormatList keywords)  /* format keywords returned */
{
  BigString buffer;     /* holds scanned format keywords */
  int cnt = 0;
  int end_ch;

  /* scan in each keyword following "format" until the keyword "tamrof"
           is encountered */
  for(;;)
  { /* read keywords until EOF or "tamrof" found */
/*    if (GSrdword(buffer, BIGSTRINGLEN) == EOF)*/
    if (rd_stringhelp(buffer, BIGSTRINGLEN, NOSPACES, &end_ch) == EOF)
    {
      GSelog(0, "Unexepected EOF while reading format keywords\n");
      return(FALSE);
    }
    if (GSstricmp(buffer, "tamrof") == 0)
    {
      *numwords = cnt;
      if (end_ch != '\n')
      {
        ungetc(skipwaste(), GSDbfp);
      }
      return(TRUE);
    }
    if (cnt == MAXFORMAT)
    {
      GSelog(0, "Too many keywords in format statement\n");
      return(FALSE);
    }
    if (strlen((char *)buffer) > MAXKWLEN)
    {
      GSelog(0, "Format statement keyword too long: |%s|\n", buffer);
      return(FALSE);
    }
    strcpy(keywords[cnt++], buffer);
  } /* for (;;) */
}
