/***********************************************************************
GeoSim pick list handling routines.
***********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "geosim.h"
#include "gspriv.h"

#define LABELOFFSET 3 /* how many pixels from label left edge to place label */

typedef struct {
  char		*wname;
  char		*daname;
  int		cols;
  int		itemx;
  int		itemy;
  int		itemht;
  int		itemwd;
  int		colgap;
  int		rowgap;
  int		currpagesize;
  int		firstitem;
  GSjustType	just;
  int		(*psize_func)(GSlistHandle, int);
  char		*(*label_func)(int);
  void		(*settext_func)(GSlistHandle, int);
  void		(*in_func)(int);
  void		(*out_func)(void);
  void		(*click_func)(int);
  int		oldslot;
  int		pagenum;
  GSColor	bgcolor;
} GSlistType;

#define	MAX_LISTS	5
static	GSlistType	*Lists[MAX_LISTS];

/* Defined in intrface.c */
extern	GSWindow	*Win;
extern	Iconfig		InfState;

/*------------------------------------------------------------------------------
  Returns TRUE if "list" is a valid list handle, else returns FALSE.
------------------------------------------------------------------------------*/
boolean validhandle (
  GSlistHandle list
)
{
  return ((list >= 0) && (list < MAX_LISTS));
}

static void fill_drag_area(char *wname, char *daname)
{
	int x, y, wd, ht;
	GSColor col;
	GSgetdragareaspecs(wname, daname, &x, &y, &wd, &ht, NULL, &col);
	GSsetcurrentnamedwindow(wname);
	GSsetcolor(col);
	GSwfillrect(x, y, wd, ht);
}

/*------------------------------------------------------------------------------
  Returns an available list handle.  Fatal error if no lists available.
------------------------------------------------------------------------------*/
GSlistHandle getlisthandle (
  void
)
{
  int i;

  for (i = 0; i < MAX_LISTS; i++)
  {
    if (Lists[i] == NULL)
    {
      Lists[i] = (GSlistType *)malloc(sizeof(GSlistType));
      if (Lists[i] == NULL)
      {
        GSelog(-1, "Out of memory for lists.\n");
      }
      else
      {
        return (i);
      }
    }
  }
  /*--- If we get this far, no free lists available ---*/
  GSelog(-1, "Reached maximum number of lists.\n");
  return 0; /* For the compiler - you'll never get here. */
}


/*------------------------------------------------------------------------------
  List destructor.
------------------------------------------------------------------------------*/
GSlistHandle GSdismantlelist (
  GSlistHandle list
)
{
  if (validhandle(list))
  {
    free((void *)Lists[list]);
    Lists[list] = NULL;
  }
  return (NO_HANDLE);
}


/*------------------------------------------------------------------------------
  Register variables for the current list.
------------------------------------------------------------------------------*/
GSlistHandle GSlistsetup (
  char	*wname,			/*--- name of window holding list ---*/
  char	*daname,		/*--- name of drag-area holding list ---*/
  int	cols,			/*--- Number of columns on page ---*/
  char	*labelfield,		/*--- Field descriptor for a label ---*/
  int	colgap,			/*--- Gap (in pixels) between columns ---*/
  int	rowgap,			/*--- Gap (in pixels) between rows ---*/
  GSjustType just,		/*-- justification of labels --*/	
  int	(*psize_func)(GSlistHandle, int),/* returns # of items on current page */
  char	*(*label_func)(int),	/*--- returns requested label ---*/
  void	(*settext_func)(GSlistHandle, int), /* sets text attributes for items */
  void	(*in_func)(int),	/*--- response when cursor in list ---*/
  void	(*out_func)(void),	/*--- response when cursor leaves list ---*/
  void	(*click_func)(int))	/*--- response when user clicks ---*/
{
  GSlistHandle	list;
  int		i, j;

  list = getlisthandle();
  Lists[list]->wname = wname;
  Lists[list]->daname = daname;
  Lists[list]->cols = cols;
  GSgetfieldrect(labelfield, &(Lists[list]->itemx), &(Lists[list]->itemy),
                 &(Lists[list]->itemwd), &(Lists[list]->itemht), NULL);
  Lists[list]->colgap = colgap;
  Lists[list]->rowgap = rowgap;
  Lists[list]->psize_func = psize_func;
  Lists[list]->label_func = label_func;
  Lists[list]->settext_func = settext_func;
  Lists[list]->in_func = in_func;
  Lists[list]->out_func = out_func;
  Lists[list]->click_func = click_func;
  i = getname(Win, sizeof(GSWindow), InfState.numwindows, wname, "window");
  j = getname(Win[i].wb, sizeof(GSButton), Win[i].numbuttons, daname,
              "dragarea");
  Lists[list]->bgcolor = Win[i].wb[j].bgcolor;
  Lists[list]->just = just;
  return (list);
}


/*------------------------------------------------------------------------------
  Display item "itemno" from a list.
------------------------------------------------------------------------------*/
void GSdrawlistitem (
  GSlistHandle	list,
  DispType	displaytype,
  int		slot
)
{
  int			xslot, yslot;	/*-- col and row in terms of list item
                                             slots --*/
  int			bx, by;		/*-- button offset within window --*/
  int			bwd, bht;	/*-- button width and height --*/
  ButtonStatusType	bstatus;	/*-- DISABLED, ENABLED, etc. --*/
  int			x, y;		/*-- Coordinate of label --*/
  static int		prev_slot = -1;
  char			*p;

  if (!validhandle(list)) {
    GSelog(-1, "Invalid GSlistHandle passed to GSdrawlistitem\n");
  }
  if ((displaytype == Unhighlight) && (Lists[list]->oldslot == -1)) {
    return;
  }
  /*--- get specs for the drag area that contains the list ---*/
  GSgetdragareaspecs(Lists[list]->wname, Lists[list]->daname, &bx, &by, &bwd,
                     &bht, &bstatus, NULL);
  if (slot == CURR_LIST_ITEM) {
    slot = Lists[list]->oldslot;
  }
  else if (slot == PREV_LIST_ITEM) {
    slot = prev_slot;
  }
  xslot = (slot - Lists[list]->firstitem) % Lists[list]->cols;
  yslot = (slot - Lists[list]->firstitem) / Lists[list]->cols;
  x = bx + Lists[list]->itemx + xslot*(Lists[list]->itemwd+Lists[list]->colgap);
  y = by + Lists[list]->itemy + yslot*(Lists[list]->itemht+Lists[list]->rowgap);
/*-----
  x = bx + Lists[list]->itemx + xslot*Lists[list]->itemwd;
  y = by + Lists[list]->itemy + yslot*Lists[list]->itemht;
-------*/

  if (/*(x >= bx) &&*/ (y >= by) /*&& (Lists[list]->itemwd + x <= bwd + bx)*/
  			&& (Lists[list]->itemht + y <= bht + by)) { /* draw inside dragarea */
	  if (displaytype == Highlight) {
	    GSsetcolor(InfState.highlight);
	    GSwfillrect (x, y, Lists[list]->itemwd, Lists[list]->itemht);
	  }
	  else {	/*-- (displaytype != Unhighlight) --*/
	    GSsetcolor(Lists[list]->bgcolor);
	    GSwfillrect (x, y, Lists[list]->itemwd, Lists[list]->itemht);
	  }
	  (*Lists[list]->settext_func)(list, slot);	/*-- set the text specs --*/
	  p = (*Lists[list]->label_func)(slot);
	  switch (Lists[list]->just) {
	    case JUST_LEFT:
	      GSwwriteclippedmsg(x+LABELOFFSET, y+SMALL,
	                         Lists[list]->itemwd-LABELOFFSET-1, Lists[list]->itemht,
	                         x+LABELOFFSET, y+SMALL, p);
	      break;
	    case JUST_RIGHT:
	      GSwwriteclippedmsg(x, y+SMALL, Lists[list]->itemwd-1,
	                         Lists[list]->itemht,
	                         x+Lists[list]->itemwd-LABELOFFSET-1-GStextwidth(p),
	                         y+SMALL, p);
	      break;
	    case JUST_CENTER:
	      GSwwriteclippedmsg(x+LABELOFFSET, y+SMALL,
	                         Lists[list]->itemwd-LABELOFFSET-1, Lists[list]->itemht,
	                         x+(Lists[list]->itemwd-GStextwidth(p))/2, y+SMALL, p);
	      break;
	  }
  }
  prev_slot = Lists[list]->oldslot;
}


/***********************************************************************
Display page "pagenum" from a list.  The first page is 1.
WARNING: it is assumed that the drag area has already been cleared.
***********************************************************************/
void GSlistpage (
  GSlistHandle	list,
  int		pagenum         /* Which page to display */
)
{
  int itemcnt;

  if (!validhandle(list)) {
    GSelog(-1, "Invalid GSlistHandle passed to GSlistpage\n");
  }
  Lists[list]->pagenum = pagenum;
  for (itemcnt=1, Lists[list]->firstitem=0; itemcnt<pagenum; itemcnt++)
    Lists[list]->firstitem += (*Lists[list]->psize_func)(list, itemcnt);
  Lists[list]->currpagesize = (*Lists[list]->psize_func)(list, pagenum);
  for (itemcnt=0; itemcnt<Lists[list]->currpagesize; itemcnt++)
  {
    Lists[list]->oldslot = itemcnt+Lists[list]->firstitem;
    GSdrawlistitem(list, New, CURR_LIST_ITEM);
  }
  Lists[list]->oldslot = -1;
}

/***********************************************************************
Display pageSize items, starting with firstItem.  JBR
Unlike GSlistpage, this function clears the dragarea.
***********************************************************************/
void GSlistfromitem(GSlistHandle list, int firstItem)
{
	int itemCnt;
	
	if (!validhandle(list))
		GSelog(-1, "Invalid GSlistHandle passed to GSlistpage\n");

	fill_drag_area(Lists[list]->wname, Lists[list]->daname);
	/* Is it OK not to set Lists[]->pageNum here? */
	Lists[list]->firstitem = firstItem;
	Lists[list]->currpagesize = (*Lists[list]->psize_func)(list, 1);
	for (itemCnt = 0; itemCnt < Lists[list]->currpagesize; itemCnt++) {
		Lists[list]->oldslot = itemCnt+Lists[list]->firstitem;
		GSdrawlistitem(list, New, CURR_LIST_ITEM);
		}
	Lists[list]->oldslot = -1;
}

/*************************************************************************
 selectitem  Selects the item in the list under relx and rely
 Called by GShandlelist when status == MOUSE_CLICK or status == DRAG_FINAL
**************************************************************************/
void selectitem (
  GSlistHandle	list,
  int		relx,
  int		rely
)
{
  int rows, slot;
  GSlistType *l = Lists[list]; /* for convenience */
  rows = (l->currpagesize+(l->cols-1))/l->cols;
  slot = rely * l->cols + relx + l->firstitem;

  /* unhighlight old label */
 /* if ((l->oldslot >= l->firstitem)
  		&& (l->oldslot < l->firstitem+l->currpagesize))*/
  	GSdrawlistitem(list, Unhighlight, CURR_LIST_ITEM);

  if ((slot - l->firstitem) >= l->currpagesize ||
  rely >= rows || relx > (l->cols-1) || (relx < 0) || (rely < 0) )
 /* added check on relx, rely JBR */
  {  GSbell();}
  else
  {
    l->oldslot = slot;
    if (*l->click_func != NULL)
      (*l->click_func)(slot);
  }
}

/*------------------------------------------------------------------------------
 processinside calls the mouseinside funct of the list for the
 item under relx and rely.
 Called by GShandlelist when status == MOUSE_INSIDE or status == DRAG_PROCESS
------------------------------------------------------------------------------*/
void processinside (
  GSlistHandle	list,
  int		relx,
  int		rely
)
{
  int rows, slot;

  rows = (Lists[list]->currpagesize+(Lists[list]->cols-1))/Lists[list]->cols;
  slot = rely * Lists[list]->cols + relx + Lists[list]->firstitem;

  if (Lists[list]->oldslot == slot) return;

  /* unhighlight old label */
  GSdrawlistitem(list, Unhighlight, CURR_LIST_ITEM);

/* added check on rely  JBR */
  if ((slot - Lists[list]->firstitem) >= Lists[list]->currpagesize ||
  rely >= rows || relx > (Lists[list]->cols-1) || relx < 0 || (rely < 0)
)
    /* Went outside the set of labels.  Treat this like being out
   of drag area. */
  {
    Lists[list]->oldslot = -1;
    if (*Lists[list]->out_func != NULL)
      (*Lists[list]->out_func)();
	return;
  }
  Lists[list]->oldslot = slot;
  GSdrawlistitem(list, Highlight, CURR_LIST_ITEM);
  if (*Lists[list]->in_func != NULL)
    (*Lists[list]->in_func)(slot);
}

/*------------------------------------------------------------------------------
  Process the mouse action within a list.
------------------------------------------------------------------------------*/
void GShandlelist (
  GSlistHandle	 list,
  int            relx,
  int            rely, 
  DragStatusType status
)
{
  if (!validhandle(list))
  {
    GSelog(-1, "Invalid GSlistHandle passed to GShandlelist\n");
  }

  if (status == MOUSE_OUTSIDE)
  { /*--- unhighlight label ---*/
    GSdrawlistitem(list, Unhighlight, CURR_LIST_ITEM);
    if  (*Lists[list]->out_func != NULL)
	    (*Lists[list]->out_func)();
    Lists[list]->oldslot = -1;
    return;
  }

  if (relx - Lists[list]->itemx < 0) {
    relx = -1;
  }
  else {
    relx = (relx - Lists[list]->itemx)/(Lists[list]->itemwd+Lists[list]->colgap);
  }
/*------------------------------------------------------------------------------
  relx = (relx - Lists[list]->itemx + Lists[list]->cols - 1)/
         (Lists[list]->itemwd+Lists[list]->colgap);
------------------------------------------------------------------------------*/
/* added this check on rely - analogous to check on relx.  JBR */
  if (rely - Lists[list]->itemy < 0) {
    rely = -1;
  }
  else {
  rely = (rely - Lists[list]->itemy) / (Lists[list]->itemht+Lists[list]->rowgap);
  }

  if (status == MOUSE_CLICK)
  {
    selectitem(list, relx, rely);
  }
  else if (status == MOUSE_INSIDE)
  {
    processinside(list, relx, rely);
  }
  else if (status == DRAG_FINAL)
  {
    selectitem(list, relx, rely);
  }
  else if (status == DRAG_PROCESS)
  {
    processinside(list, relx, rely);
  }
}
