#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "geosim.h"
#include "gspriv.h"

#define TAG4 0x844e5552L  /* ID tag: 'RUN' + 0x04000000 + 0x80000000L */
#define TAG8 0x884e5552L  /* ID tag: 'RUN' + 0x08000000 + 0x80000000L */
#define TAG12 0x8c4e5552L /* ID tag: 'RUN' + 0x0c000000 + 0x80000000L */
#define TAG16 0x904e5552L /* ID tag: 'RUN' + 0x10000000 + 0x80000000L */
#define MAXPIX 640  /* Maximum image width */
#define fread_byte(p, f)     fread(p, 1, 1, f)

/****************************************************************************
****************************************************************************/

static unsigned short twiddled_short(void *p)
{
    unsigned char    *c = (unsigned char *) p;
    unsigned short    lo, hi;

    lo = c[0];
    hi = c[1];
    return (hi << 8) | lo;
}

static size_t fread_short(void *ptr, FILE *file)
{
    char    t, *c;

    if (fread(ptr, 2, 1, file) != 1)  return 0;

    if (GSbigendian())
    {
        c = (char *) ptr;
        t = c[0];  c[0] = c[1];  c[1] = t;
    }

    return 1;
}


static size_t fread_long(void *ptr, FILE *file)
{
    char    t, *c;

    if (fread(ptr, 4, 1, file) != 1)  return 0;

    if (GSbigendian())
    {
        c = (char *) ptr;
        t = c[0];  c[0] = c[3];  c[3] = t;
        t = c[1];  c[1] = c[2];  c[2] = t;
    }

    return 1;
}

size_t twiddle_long(void *ptr)
{
	char t, *c;
	
	if (GSbigendian()) {
        c = (char *) ptr;
        t = c[0];  c[0] = c[3];  c[3] = t;
        t = c[1];  c[1] = c[2];  c[2] = t;
    }

    return 1;
}


/***********************************************************************
Read a runlength encoded image from file, and display the requested
part on the screen.
***********************************************************************/
void GSwdrawfilepict(FILE *fp, int x, int y, int wd, int ht,
                    int offx, int offy, 
                    GSColor (*user_func)(unsigned short))
/*, int, int, int))*/
{
  /*char charbuff[MAXPIX];*/
  GSColor charbuff[MAXPIX];
  unsigned long tag;
  long bytecnt;
  short count;
  int i, j;
  unsigned short val, srun, rows, cols;
  unsigned char crun;
  unsigned int length, currcol;

  /* Initialize in case picture smaller than spec. */
  memset(charbuff, -1, sizeof(charbuff));
  /* Seek to beginning of file, just to be safe */
  rewind(fp);

  if (fread_long(&tag, fp) != 1)  GSelog(-1, "Unable to read tag\n");
  if (fread_short(&rows, fp) != 1)  GSelog(-1, "Unable to read rows\n");
  if (fread_short(&cols, fp) != 1)  GSelog(-1, "Unable to read cols\n");
  if ((rows < offy) || (cols < offx))
    return;  /* Doesn't want any part of this picture */
  if (cols > MAXPIX)
    GSelog(-1, "Too many columns (%d) for picture\n", cols);
  for (i=0, bytecnt=0; i<=offy; i++)
  {
    fread_short(&count, fp);
    bytecnt += count;
  }
  fseek(fp, bytecnt, SEEK_SET); /* Seek to beginning of desired row */
  ht = MIN(ht, rows-offy); /* Can only do as many rows as there are */
  for (i=0; i<ht; i++)
  { /* For each row */
    for (currcol=0; currcol < cols;)
    {
      switch (tag) {
        case TAG4:
          fread_byte(&crun, fp);
          length = crun >> 4;
          if (length == 0) length = 16;
	  val = crun & 0xf;
          break;

        case TAG8:
          fread_short(&srun, fp);
          length = srun & 0xff;
          if (length == 0) length = 256;
          val = srun >> 8;
          break;

        case TAG12:
          fread_short(&srun, fp);
          length = srun >> 12;
          if (length == 0) length = 16;
          val = srun &0xfff;
          break;

        case TAG16:
          fread_byte(&crun, fp);
          length = crun;
          if (length == 0) length = 256;
          fread_short(&val, fp);
          break;

        default:
          GSelog(-1, "Unknown tag type (%lx)\n", tag);
      } /* Switch */
      if (user_func != NULL)
        val = (*user_func)(val);
      for (j=0; j<length; j++, currcol++)
        charbuff[currcol] = (unsigned char)val;
    }
    GSwfastrow(x, y+i, &charbuff[offx], wd);
  }
}


/***********************************************************************
Read a runlength encoded image into memory for future processing.
fp is a pointer to the file from which to read the image,
pict is the structure describing the picture,
and mem is the memory area provided in which to store the image.
This routine relies on the caller to pass adequate space in mem --
at least (filesize - 8) bytes.
Returns TRUE iff processing the image was successful.
***********************************************************************/
boolean GSinitpic(FILE *fp, GSPicture *pict, void *mem)
{
  long filesize;  /* Size of file in bytes */
  int  i;

  /* Seek to beginning of file, just to be safe */
  rewind(fp);
  if (fread_long(&pict->tag, fp) != 1)  return(FALSE);
  if ((pict->tag != TAG4) && (pict->tag != TAG8) &&
      (pict->tag != TAG12) && (pict->tag != TAG16))
    return(FALSE);
  if (fread_short(&pict->rows, fp) != 1)  return(FALSE);
  if (fread_short(&pict->cols, fp) != 1)  return(FALSE);
  filesize = GSfilesize(fp);
  if (fread(mem, filesize-8, 1, fp) != 1)  return(FALSE);

  pict->rowvec = (unsigned short *)mem;
  pict->rowvec[0] = 0; /* Don't need to skip file header when in memory */
  for (i = 1;  i <= pict->rows;  i++)
    pict->rowvec[i] = twiddled_short(pict->rowvec + i);
  pict->codes = (unsigned char *)mem + 2*(pict->rows+1);
  switch (pict->tag) {
    case TAG4: pict->runsize = 1; break;
    case TAG8: pict->runsize = 2; break;
    case TAG12: pict->runsize = 2; break;
    case TAG16: pict->runsize = 3; break;
  }
  return(TRUE);
}


/***********************************************************************
Return the length and value of the next runlength code.
***********************************************************************/
static void getnextrun(unsigned long tag, unsigned char *ptr,
                unsigned int *runlength, unsigned short *val)
{
  switch (tag) {
    case TAG4:
      *runlength = *ptr >> 4;
      if (*runlength == 0) *runlength = 16;
      *val = *ptr & 0x0F;
      break;

    case TAG8:
      *runlength = *ptr;
      if (*runlength == 0) *runlength = 256;
      *val = ptr[1];
      break;

    case TAG12:
/*      *runlength = *(unsigned short *)ptr >> 12;*/
      *runlength = twiddled_short(ptr) >> 12;
      if (*runlength == 0) *runlength = 16;
/*      *val = *(unsigned short *)ptr &0xfff;*/
      *val = twiddled_short(ptr) & 0x0FFF;
      break;

    case TAG16:
      *runlength = *ptr;
      if (*runlength == 0) *runlength = 256;
/*      *val = *(unsigned short *)(ptr+1);*/
      *val = twiddled_short(ptr+1);
      break;
  } /* end switch (tag) */
}


/*------------------------------------------------------------------------------
  Return the number of rows and/or the number of columns in a runlength
  encoded file.  Assumes that fp has been properly initialized.
------------------------------------------------------------------------------*/
void GSgetfilepictspecs (
  FILE			*fp,
  unsigned short	*n_rows_p,
  unsigned short	*n_cols_p
)
{
  unsigned long	tag;

  rewind(fp);	/*-- Seek to beginning of file, just to be safe --*/
  if (fread_long(&tag, fp) != 1) {	/*-- read past the tag --*/
    GSelog(-1, "Unable to read tag\n");
  }
  if (fread_short(n_rows_p, fp) != 1) {
    GSelog(-1, "Unable to read rows\n");
  }
  if (fread_short(n_cols_p, fp) != 1) {
    GSelog(-1, "Unable to read cols\n");
  }
}


/*------------------------------------------------------------------------------
  Return in "buff" "n" picture values from the row beginning at ("x", "y")
  in picture file "fp".  Assumes that fp has been properly initialized.
------------------------------------------------------------------------------*/
void GSgetfilepixel (
  FILE			*fp,
  int			x,
  int			y,
  unsigned short	*buff,
  int			n
)
{
  unsigned long		tag;
  long			bytecnt;
  short			count;
  int			i, j;
  unsigned short	val, srun, rows, cols;
  unsigned char		crun;
  unsigned int		length, currcol;

  /*-- Seek to beginning of file, just to be safe --*/
  rewind(fp);
  if (fread_long(&tag, fp) != 1) {
    GSelog(-1, "Unable to read tag\n");
  }
  if (fread_short(&rows, fp) != 1) {
    GSelog(-1, "Unable to read rows\n");
  }
  if (fread_short(&cols, fp) != 1) {
    GSelog(-1, "Unable to read cols\n");
  }
  if ((rows < y) || (cols < x)) {
    return;  /*-- Doesn't want any part of this picture --*/
  }
  for (i = 0, bytecnt = 0; i <= y; i++) {
    /*-- read row vectors to find offset of desired row --*/
    fread_short(&count, fp);
    bytecnt += count;
  }
  if (fseek(fp, bytecnt, SEEK_SET)) { /*-- Seek to offset of desired row --*/
    GSelog(-1, "Cannot seek to offset %ld in GSgetfilepixel\n");
  }
  length = 0;
  for (currcol = 0; currcol + length < x; ) {
    currcol += length;
    switch (tag) {
      case TAG4:
        fread_byte(&crun, fp);
        length = crun >> 4;
        if (length == 0) {
          length = 16;
	}
        val = crun & 0xf;
        break;
      case TAG8:
        fread_short(&srun, fp);
        length = srun & 0xff;
        if (length == 0) {
          length = 256;
	}
        val = srun >> 8;
        break;
      case TAG12:
        fread_short(&srun, fp);
        length = srun >> 12;
        if (length == 0) {
          length = 16;
	}
        val = srun &0xfff;
        break;
      case TAG16:
        fread_byte(&crun, fp);
        length = crun;
        if (length == 0) {
          length = 256;
	}
        fread_short(&val, fp);
        break;
    }
  }
  length -= x - currcol;
  for (currcol = 0 ; currcol < MIN(n, cols); ) {
    for (j=0; j<length; j++, currcol++) {
      if (currcol >= n) {
        return;
      }
      buff[currcol] = val;
    }
    switch (tag) {
      case TAG4:
        fread_byte(&crun, fp);
        length = crun >> 4;
        if (length == 0) {
          length = 16;
	}
        val = crun & 0xf;
        break;
      case TAG8:
        fread_short(&srun, fp);
        length = srun & 0xff;
        if (length == 0) {
          length = 256;
	}
        val = srun >> 8;
        break;
      case TAG12:
        fread_short(&srun, fp);
        length = srun >> 12;
        if (length == 0) {
          length = 16;
	}
        val = srun &0xfff;
        break;
      case TAG16:
        fread_byte(&crun, fp);
        length = crun;
        if (length == 0) {
          length = 256;
	}
        fread_short(&val, fp);
        break;
    }
  }
}

/***********************************************************************
Return in "buff" "n" picture values from the row beginning at ("x", "y")
in picture "pic".
***********************************************************************/
void GSgetpixel(GSPicture *pic, int x, int y, unsigned short *buff, int n)
{
  unsigned char *ptr;  /* Pointer to current runlength code */
  long offset = 0;
  int i, j, currcount = 0, curroutcol = 0;
  unsigned short val;
  unsigned int runlength;

  if ((x >= pic->cols) || (y >= pic->rows))
    for (i=0; i<n; i++)
      buff[i] = (unsigned short)-1;
  for(i=1; i<=y; i++)
    offset += pic->rowvec[i];
  ptr = pic->codes + offset - pic->runsize;
  runlength = 0;
  while (currcount + runlength <= x)
  {
    currcount += runlength;
    ptr += pic->runsize;
    getnextrun(pic->tag, ptr, &runlength, &val);
  }
  runlength -= x-currcount;
  currcount = x;
  while ((curroutcol < n) && (currcount < pic->cols))
  {
    for (j=0; j<runlength; j++)
      if (curroutcol < n)
      {
        buff[curroutcol++] = val;
        currcount++;
      }
    ptr += pic->runsize;
    getnextrun(pic->tag,ptr, &runlength, &val);
  }
  for (; curroutcol<n; curroutcol++) /* Fill out rest of row if necessary */
    buff[curroutcol] = (unsigned short)-1;
}


/***********************************************************************
Draw the requested part of an in-memory runlength encoded image
onto the screen.
"pic" is the descriptor for the image.
(offx, offy) is the upper left corner of the input image, in terms of
its (unscaled) pixels.
(x, y, wd, ht) is the output block in terms of screen pixels.
"scale" determines the size of an input pixel in terms of screen pixels.
"user_func" converts an image pixel into a screen color.
***********************************************************************/
void GSwdrawpict(GSPicture *pic, int x, int y, int wd, int ht,
	   int offx, int offy, int scale,
           GSColor (*user_func)(unsigned short, int, int, int))
{
  unsigned char *ptr,
                *srptr; /*--- saves place at start of row for repetitions ---*/
  unsigned short runval,
                 val;
  unsigned int   runlength;
  int            currcol; /* Current column in the input image */
  int            curroutcol; /* Current (scaled) column in the output buffer */
  int            i,
                 j,
                 k;
  int            inrow; /* Current row in the picture to be processed */
  GSColor charbuff[MAXPIX];
 /* char           charbuff[MAXPIX];*/
  long           offset; /* Offset for current row from start of code list */

  if ((pic->rows < offy) || (pic->cols < offx))
    return;  /* Doesn't want any part of this picture */
  /* Initialize in case picture smaller than spec. */
  memset(charbuff, -1, sizeof(charbuff));
  for (i=1, offset=0; i<offy; i++)
    offset += pic->rowvec[i];
  inrow = offy;
  ht = MIN(ht, (pic->rows-offy)*scale); /* Can only do as many rows as there are */
  for (i=0; i<ht; i++)
  { /* Do a row */
    if ((scale != 1) && (i % scale))  /* Repeat current row */
    {
      ptr = srptr;
      /*GSwfastrow(x, y+i, charbuff, wd);*/
    }
    else
    { /* Build a new row */
      offset += pic->rowvec[inrow++];
      ptr = pic->codes + offset - pic->runsize;
      srptr = ptr;
    }
      currcol = 0;
      runlength = 0;
      while (currcol + runlength <= offx)
      { /* First, skip the part below the x offset */
        currcol += runlength;
        ptr += pic->runsize;
        getnextrun(pic->tag, ptr, &runlength, &runval);
      }
      runlength -= offx-currcol;
      currcol = offx;
      curroutcol = 0;
      while ((curroutcol < wd) && (currcol < pic->cols))
      {
/*        if (user_func != NULL)
          val = user_func(runval, curroutcol, i);*/
        for (j=0; j<runlength; j++)
        {
          for (k=0; k<scale; k++)
            if (curroutcol<wd)
            {
              if (user_func != NULL)
                val = user_func(runval, curroutcol, i/scale, k);
              charbuff[curroutcol++] = val;
	    }
          currcol++;
	}
	ptr += pic->runsize;
        getnextrun(pic->tag, ptr, &runlength, &runval);
      }
      GSwfastrow(x, y+i, charbuff, wd);
    /*}*/ /* else */
  } /* for */
}
